from pyVmomi import vim
from pyVmomi import vmodl

from .views import get_all_vms


def get_field_by_key(content, field_key):
    """
    Get a custom field by its key
    :param content: Service Content
    :type content: vim.ServiceContent
    :param field_id: Key of the field
    :type field_id: str or unicode
    :return: The requested field, or None if the field does not exist
    :rtype: vim.CustomFieldDef or NoneType
    """
    fields_manager = content.customFieldsManager
    for field in fields_manager.field:
        if hasattr(field, "key") and field.key == field_key:
            return field


def get_field_by_name(content, fieldname):
    """
    Get a custom field by its name
    :param content: Service Content
    :type content: vim.ServiceContent
    :param fieldname: Name of the field
    :type fieldname: str or unicode
    :return: The requested field, or None if the field does not exist
    :rtype: vim.CustomFieldDef or NoneType
    """
    fields_manager = content.customFieldsManager
    for field in fields_manager.field:
        if hasattr(field, "name") and field.name == fieldname:
            return field


def get_or_create_custom_field(
    content, fieldname, fieldtype=vim.VirtualMachine
):
    """
    Create a new custom field
    :param content: Service Content
    :type content: vim.ServiceContent
    :param fieldname: Name of the new field
    :type fieldname: str or unicode
    :param fieldtype: Type of the new Field (vim.VirtualMachine,
    vim.HostSystem etc.)
    :type fieldtype: object
    :return: The newly created field
    :rtype: vim.CustomFieldDef
    """
    fields_manager = content.customFieldsManager
    # Check if the field already exists
    field = get_field_by_name(content, fieldname)
    # Return the field if it exists
    if field:
        return field
    return fields_manager.AddCustomFieldDef(
        name=fieldname,
        moType=fieldtype
        # fieldDefPolicy=None,
        # fieldPolicy=None
    )


def set_field(content, vm, name, value):
    """
    Set a custom field for a given VM
    :param content: Service Content
    :type content: vim.ServiceContent
    :param vm: Virtual Machine whose field is to be set
    :type vm: vim.VirtualMachine
    :param name: Name of the custom field
    :type name: str or unicode
    :param value: Value of the custom field
    :type value: str or unicode
    """
    # FIXME There should be a way to get the field name from vm.customValue
    #       -> Get rid of the content parameter
    # Check if the field is already set to the desired value
    for cfield in vm.customValue:
        field = get_field_by_key(content, cfield.key)
        if field.name == name:
            # Only set the value if it differs
            if cfield.value == value:
                return
            break
    # Field was not found or value differs, set its value
    # FIXME Passing the name instead of the key is weird but works
    vm.setCustomValue(key=name, value=value)


def clear_field(content, vm, field):
    """
    Clear a field by setting its value to empty string
    :param content: Service Content
    :type content: vim.ServiceContent
    :param vm: Virtual Machine whose field is to be set
    :type vm: vim.VirtualMachine
    :param name: Name of the custom field
    :type name: str or unicode
    :param value: Value of the custom field
    :type value: str or unicode
    """
    return set_field(content, vm, field, "")


def find_vms_by_custom_field(content, name, value):
    """
    Find virtual machines who match a certain criteria
    :param content: Service Content
    :type content: vim.ServiceContent
    :param name: Name of the annotation/custom field (not the integer key)
    :type name: str or unicode
    :param value: Value of the annotation/custom field
    :type value: str or unicode
    :return: A list of VMs that match
    :rtype: list
    """
    matching = []
    vm_list = get_all_vms(content)
    for vm in vm_list:
        for cfield in vm.customValue:
            field = get_field_by_key(content, cfield.key)
            if field.name == name and cfield.value == value:
                matching.append(vm)
    return matching


def get_obj_field_value(content, obj, name):
    """
    Get the value of the named field for a given object
    :param content: Service Content
    :type content: vim.ServiceContent
    :param name: Name of the annotation/custom field (not the integer key)
    :type name: str or unicode
    :return: The value of the field
    :rtype: str or NoneType
    """
    field = get_field_by_name(content, name)
    if not field:
        # Named field does not exist
        return
    fld_key = field.key
    for fld in obj.customValue:
        if fld_key == fld.key:
            return fld.value



# coding: utf-8

import requests



def get_tag_by_name(stub, tag_name):
    """
    Get a TagModel bbyy its name
    """
    import com.vmware.cis.tagging_client as tagclient

    tag_svc = tagclient.Tag(stub)
    all_tag_ids = tag_svc.list()
    for tag_id in all_tag_ids:
        tag_obj = tag_svc.get(tag_id)
        if tag_obj.name == tag_name:
            return tag_obj


def tag_vm(stub, vm, tag):
    """
    Tag a VM
    """
    import com.vmware.cis.tagging_client as tagclient
    from com.vmware.vapi.std_client import DynamicID

    if isinstance(tag, tagclient.TagModel):
        tag_id = tag.id
    else:
        tag_obj = get_tag_by_name(stub, tag)
        tag_id = tag_obj.id
    if isinstance(vm, DynamicID):
        obj_did = vm
    else:
        obj_did = DynamicID(type="VirtualMachine", id=vm._moId)
    tag_ass = tagclient.TagAssociation(stub)
    result = tag_ass.attach(tag_id=tag_id, object_id=obj_did)
    return result



from .views import (get_all_clusters, get_vm_by_name)
from pyVmomi import vim
from pyVim.task import WaitForTask


def get_cluster_affinity_groups(cluster):
    """
    Get the VM/Host groups of a given cluster
    https://github.com/vmware/pyvmomi/issues/663#issuecomment-374397624
    """
    return cluster.configurationEx.group


def get_cluster_affinity_rules(cluster):
    """
    Get the VM/Host rules of a given cluster
    """
    # return cluster.configurationEx.rule
    return cluster.configuration.rule


def get_affinity_rules(cluster, search_criteria):
    """
    Get an affinity rule by its name
    """
    rules = []
    for rule in cluster.configuration.rule:
        # Keys are unique
        if rule.key == search_criteria:
            return [rule]
        # Names are not
        if rule.name == search_criteria:
            rules.append(rule)
    return rules if rules else None


def get_affinity_rules_affecting(content, vm):
    """
    Get a list of rules affecting a VM
    :param content: VMware API content
    :type content: pyVmomi.VmomiSupport.vim.ServiceInstanceContent
    :param vm: Name of the VM (or the API object)
    :type vm: str or pyVmomi.VmomiSupport.vim.VirtualMachine
    :return: A list of dicts holding the cluster and the rules affecting the VM
    :rtype: list
    """
    if isinstance(vm, str):
        vm = get_vm_by_name(content, vm)
    rules = []
    for cl in get_all_clusters(content):
        for rule in cl.configuration.rule:
            if vm in rule.vm:
                rules.append({"cluster": cl, "rule": rule})
    return rules if rules else None


def create_affinity_rule(cluster, vms, name=None, together=True):
    """
    Create a new affinity rule for VMs
    https://github.com/vmware/pyvmomi-community-samples/issues/247
    """
    if not name:
        name = 'affinity_shortmomi'
    if together:
        aff_rule_spec = vim.cluster.AffinityRuleSpec
    else:
        aff_rule_spec = vim.cluster.AntiAffinityRuleSpec
    rule = aff_rule_spec(vm=vms, enabled=True, mandatory=True, name=name)
    rule_spec = vim.cluster.RuleSpec(info=rule, operation='add')
    config_spec = vim.cluster.ConfigSpecEx(rulesSpec=[rule_spec])
    return WaitForTask(cluster.ReconfigureEx(config_spec, modify=True))


def delete_affinity_rule(cluster, rule):
    """
    Delete an affinity rule
    """
    rules = get_affinity_rules(cluster, rule)
    if not rules:
        # Nothing to do
        return

    spec = vim.cluster.ConfigSpecEx()
    spec.rulesSpec = []
    for rule in rules:
        rule_spec = vim.cluster.RuleSpec(operation="remove", removeKey=rule.key)
        spec.rulesSpec.append(rule_spec)
    return WaitForTask(cluster.ReconfigureEx(spec, modify=True))



from __future__ import absolute_import

from .affinity_rules import *
from .connect import (connect, ConnectionError, stub_connect)
from .custom_attributes import *
from .datastores import *
from .tasks import *
from .tags import *
from .views import *
from .vm import *



# coding: utf-8

import time

from pyVmomi import vim


def shutdown_vm(vm, blocking=True):
    if vm.runtime.powerState == "poweredOn":
        vm.ShutdownGuest()
        if blocking:
            while vm.runtime.powerState == "poweredOn":
                time.sleep(1)


# https://github.com/Akasurde/ansible-reproducers/blob/master/790/get_all_vms.py
def get_folder_path(obj):
    paths = []
    if isinstance(obj, vim.Folder):
        paths.append(obj.name)

    thisobj = obj
    while hasattr(thisobj, 'parent'):
        thisobj = thisobj.parent
        try:
            moid = thisobj._moId
        except AttributeError:
            moid = None
        if moid in ['group-d1', 'ha-folder-root']:
            break
        if isinstance(thisobj, vim.Folder):
            paths.append(thisobj.name)

    paths.reverse()
    return '/'.join(paths)



# coding: utf-8

from pyVmomi import vim
from pyVmomi import vmodl


def wait_for_tasks(service, tasks):
    """
    Given the service instance and tasks, it returns after all the
    tasks are complete
    """

    pc = service.propertyCollector

    if not isinstance(tasks, list):
        tasks = [tasks]
    taskList = [str(task) for task in tasks]

    # Create filter
    objSpecs = [vmodl.query.PropertyCollector.ObjectSpec(obj=task) for task in tasks]
    propSpec = vmodl.query.PropertyCollector.PropertySpec(
        type=vim.Task, pathSet=[], all=True
    )
    filterSpec = vmodl.query.PropertyCollector.FilterSpec()
    filterSpec.objectSet = objSpecs
    filterSpec.propSet = [propSpec]
    filter = pc.CreateFilter(filterSpec, True)

    try:
        version, state = None, None

        # Loop looking for updates till the state moves to a completed state.
        while len(taskList):
            update = pc.WaitForUpdates(version)
            for filterSet in update.filterSet:
                for objSet in filterSet.objectSet:
                    task = objSet.obj
                    for change in objSet.changeSet:
                        if change.name == "info":
                            state = change.val.state
                        elif change.name == "info.state":
                            state = change.val
                        else:
                            continue

                        if not str(task) in taskList:
                            continue

                        if state == vim.TaskInfo.State.success:
                            # Remove task from taskList
                            taskList.remove(str(task))
                        elif state == vim.TaskInfo.State.error:
                            raise task.info.error
            # Move to next version
            version = update.version
    finally:
        if filter:
            filter.Destroy()



# coding: utf-8

from __future__ import absolute_import
from pyVmomi import vim
from .tasks import wait_for_tasks


def datastore_search_file(service, datastore, filename):
    details = vim.host.DatastoreBrowser.FileInfo.Details(
        fileType=True, fileSize=True, fileOwner=True, modification=True
    )
    search = vim.host.DatastoreBrowser.SearchSpec(
        sortFoldersFirst=True, details=details, matchPattern=[filename]
    )
    path = f"[{datastore.name}]"
    task = datastore.browser.SearchSubFolders(datastorePath=path, searchSpec=search)
    wait_for_tasks(service, [task])
    results = []
    for r in task.info.result:
        for f in r.file:
            # discard folders
            if r.folderPath == f"[{datastore.name}]":
                continue
            results.append((r.folderPath, f))

    return results
